@@ -11,11 +11,11 @@
 enum {
 	ATOMIC_ALL_STEPS,
 	ATOMIC_STEP_SYSCORE_RESUME,
-	ATOMIC_STEP_SYSDEV_RESUME,
 	ATOMIC_STEP_IRQS,
 	ATOMIC_STEP_CPU_HOTPLUG,
 	ATOMIC_STEP_PLATFORM_FINISH,
 	ATOMIC_STEP_DEVICE_RESUME,
+	ATOMIC_STEP_DPM_COMPLETE,
 	ATOMIC_STEP_PLATFORM_END,
 };
 
