@@ -1,11 +1,9 @@
 /*
- * Copyright (C) 2004-2007 Nigel Cunningham (nigel at tuxonice net)
+ * Copyright (C) 2004-2010 Nigel Cunningham (nigel at tuxonice net)
  *
  * This file is released under the GPLv2.
  */
-#include <linux/module.h>
 #include <linux/resume-trace.h>
-#include <linux/syscalls.h>
 #include <linux/kernel.h>
 #include <linux/swap.h>
 #include <linux/syscalls.h>
@@ -17,12 +15,10 @@
 #include <linux/tty.h>
 #include <linux/crypto.h>
 #include <linux/cpu.h>
-#include <linux/dyn_pageflags.h>
 #include <linux/ctype.h>
 #include "tuxonice_io.h"
 #include "tuxonice.h"
 #include "tuxonice_extent.h"
-#include "tuxonice_block_io.h"
 #include "tuxonice_netlink.h"
 #include "tuxonice_prepare_image.h"
 #include "tuxonice_ui.h"
@@ -30,6 +26,11 @@
 #include "tuxonice_pagedir.h"
 #include "tuxonice_modules.h"
 #include "tuxonice_builtin.h"
+#include "tuxonice_power_off.h"
+#include "tuxonice_alloc.h"
+
+unsigned long toi_bootflags_mask;
+EXPORT_SYMBOL_GPL(toi_bootflags_mask);
 
 /*
  * Highmem related functions (x86 only).
@@ -100,7 +101,7 @@ static void copyback_high(void)
 }
 
 #else /* CONFIG_HIGHMEM */
-void copyback_high(void) { }
+static void copyback_high(void) { }
 #endif
 
 char toi_wait_for_keypress_dev_console(int timeout)
@@ -163,128 +164,51 @@ out_close:
 
 	return key;
 }
+EXPORT_SYMBOL_GPL(toi_wait_for_keypress_dev_console);
 
-struct toi_boot_kernel_data toi_bkd __nosavedata 
+struct toi_boot_kernel_data toi_bkd __nosavedata
 		__attribute__((aligned(PAGE_SIZE))) = {
-        MY_BOOT_KERNEL_DATA_VERSION,
+	MY_BOOT_KERNEL_DATA_VERSION,
 	0,
 #ifdef CONFIG_TOI_REPLACE_SWSUSP
-        (1 << TOI_REPLACE_SWSUSP) |
+	(1 << TOI_REPLACE_SWSUSP) |
 #endif
-        (1 << TOI_PAGESET2_FULL) | (1 << TOI_LATE_CPU_HOTPLUG),
+	(1 << TOI_NO_FLUSHER_THREAD) |
+	(1 << TOI_PAGESET2_FULL) | (1 << TOI_LATE_CPU_HOTPLUG),
 };
 EXPORT_SYMBOL_GPL(toi_bkd);
 
-struct block_device *toi_open_by_devnum(dev_t dev, unsigned mode)
+struct block_device *toi_open_by_devnum(dev_t dev)
 {
 	struct block_device *bdev = bdget(dev);
 	int err = -ENOMEM;
-	int flags = mode & FMODE_WRITE ? O_RDWR : O_RDONLY;
-	flags |= O_NONBLOCK;
 	if (bdev)
-		err = blkdev_get(bdev, mode, flags);
+		err = blkdev_get(bdev, FMODE_READ | FMODE_NDELAY, NULL);
 	return err ? ERR_PTR(err) : bdev;
 }
 EXPORT_SYMBOL_GPL(toi_open_by_devnum);
 
-EXPORT_SYMBOL_GPL(toi_wait_for_keypress_dev_console);
-
-extern struct hibernation_ops *hibernation_ops;
-EXPORT_SYMBOL_GPL(hibernation_ops);
-EXPORT_SYMBOL_GPL(hibernation_platform_enter);
-
-#ifdef CONFIG_TOI_CORE_EXPORTS
-#ifdef CONFIG_X86_64
-EXPORT_SYMBOL_GPL(restore_processor_state);
-EXPORT_SYMBOL_GPL(save_processor_state);
-#endif
-
-EXPORT_SYMBOL_GPL(pm_chain_head);
-EXPORT_SYMBOL_GPL(kernel_shutdown_prepare);
-EXPORT_SYMBOL_GPL(drop_pagecache);
-EXPORT_SYMBOL_GPL(restore_pblist);
-EXPORT_SYMBOL_GPL(pm_mutex);
-EXPORT_SYMBOL_GPL(pm_restore_console);
-EXPORT_SYMBOL_GPL(super_blocks);
-EXPORT_SYMBOL_GPL(next_zone);
-
-EXPORT_SYMBOL_GPL(freeze_processes);
-EXPORT_SYMBOL_GPL(thaw_processes);
-EXPORT_SYMBOL_GPL(thaw_kernel_threads);
-EXPORT_SYMBOL_GPL(shrink_all_memory);
-EXPORT_SYMBOL_GPL(shrink_one_zone);
-EXPORT_SYMBOL_GPL(saveable_page);
-EXPORT_SYMBOL_GPL(swsusp_arch_suspend);
-EXPORT_SYMBOL_GPL(swsusp_arch_resume);
-EXPORT_SYMBOL_GPL(pm_ops);
-EXPORT_SYMBOL_GPL(pm_prepare_console);
-EXPORT_SYMBOL_GPL(follow_page);
-EXPORT_SYMBOL_GPL(machine_halt);
-EXPORT_SYMBOL_GPL(block_dump);
-EXPORT_SYMBOL_GPL(unlink_lru_lists);
-EXPORT_SYMBOL_GPL(relink_lru_lists);
-EXPORT_SYMBOL_GPL(power_subsys);
-EXPORT_SYMBOL_GPL(machine_power_off);
-EXPORT_SYMBOL_GPL(suspend_devices_and_enter);
-EXPORT_SYMBOL_GPL(first_online_pgdat);
-EXPORT_SYMBOL_GPL(next_online_pgdat);
-EXPORT_SYMBOL_GPL(machine_restart);
-EXPORT_SYMBOL_GPL(saved_command_line);
-EXPORT_SYMBOL_GPL(tasklist_lock);
-#ifdef CONFIG_PM_SLEEP_SMP
-EXPORT_SYMBOL_GPL(disable_nonboot_cpus);
-EXPORT_SYMBOL_GPL(enable_nonboot_cpus);
-#endif
-#endif
+/**
+ * toi_close_bdev: Close a swap bdev.
+ *
+ * int: The swap entry number to close.
+ */
+void toi_close_bdev(struct block_device *bdev)
+{
+	blkdev_put(bdev, FMODE_READ | FMODE_NDELAY);
+}
+EXPORT_SYMBOL_GPL(toi_close_bdev);
 
 int toi_wait = CONFIG_TOI_DEFAULT_WAIT;
-
-#ifdef CONFIG_TOI_USERUI_EXPORTS
-EXPORT_SYMBOL_GPL(kmsg_redirect);
-#endif
 EXPORT_SYMBOL_GPL(toi_wait);
 
-#if defined(CONFIG_TOI_USERUI_EXPORTS) || defined(CONFIG_TOI_CORE_EXPORTS)
-EXPORT_SYMBOL_GPL(console_printk);
-#endif
-#ifdef CONFIG_TOI_SWAP_EXPORTS	/* TuxOnIce swap specific */
-EXPORT_SYMBOL_GPL(sys_swapon);
-EXPORT_SYMBOL_GPL(sys_swapoff);
-EXPORT_SYMBOL_GPL(si_swapinfo);
-EXPORT_SYMBOL_GPL(map_swap_page);
-EXPORT_SYMBOL_GPL(get_swap_page);
-EXPORT_SYMBOL_GPL(swap_free);
-EXPORT_SYMBOL_GPL(get_swap_info_struct);
-#endif
-
-#ifdef CONFIG_TOI_FILE_EXPORTS
-/* TuxOnice file allocator specific support */
-EXPORT_SYMBOL_GPL(sys_unlink);
-EXPORT_SYMBOL_GPL(sys_mknod);
-#endif
-
-/* Swap or file */
-#if defined(CONFIG_TOI_FILE_EXPORTS) || defined(CONFIG_TOI_SWAP_EXPORTS)
-EXPORT_SYMBOL_GPL(bio_set_pages_dirty);
-EXPORT_SYMBOL_GPL(name_to_dev_t);
-#endif
-
-#if defined(CONFIG_TOI_FILE_EXPORTS) || defined(CONFIG_TOI_SWAP_EXPORTS) || \
-	defined(CONFIG_TOI_CORE_EXPORTS)
-EXPORT_SYMBOL_GPL(resume_file);
-#endif
 struct toi_core_fns *toi_core_fns;
 EXPORT_SYMBOL_GPL(toi_core_fns);
 
-DECLARE_DYN_PAGEFLAGS(pageset1_map);
-DECLARE_DYN_PAGEFLAGS(pageset1_copy_map);
-EXPORT_SYMBOL_GPL(pageset1_map);
-EXPORT_SYMBOL_GPL(pageset1_copy_map);
-
 unsigned long toi_result;
-struct pagedir pagedir1 = {1};
-
 EXPORT_SYMBOL_GPL(toi_result);
+
+struct pagedir pagedir1 = {1};
 EXPORT_SYMBOL_GPL(pagedir1);
 
 unsigned long toi_get_nonconflicting_page(void)
@@ -297,16 +221,35 @@ int toi_post_context_save(void)
 	return toi_core_fns->post_context_save();
 }
 
-int toi_try_hibernate(int have_pmsem)
+int try_tuxonice_hibernate(void)
 {
 	if (!toi_core_fns)
 		return -ENODEV;
 
-	return toi_core_fns->try_hibernate(have_pmsem);
+	return toi_core_fns->try_hibernate();
 }
 
-void toi_try_resume(void)
+static int num_resume_calls;
+#ifdef CONFIG_TOI_IGNORE_LATE_INITCALL
+static int ignore_late_initcall = 1;
+#else
+static int ignore_late_initcall;
+#endif
+
+int toi_translate_err_default = TOI_CONTINUE_REQ;
+EXPORT_SYMBOL_GPL(toi_translate_err_default);
+
+void try_tuxonice_resume(void)
 {
+	/* Don't let it wrap around eventually */
+	if (num_resume_calls < 2)
+		num_resume_calls++;
+
+	if (num_resume_calls == 1 && ignore_late_initcall) {
+		printk(KERN_INFO "TuxOnIce: Ignoring late initcall, as requested.\n");
+		return;
+	}
+
 	if (toi_core_fns)
 		toi_core_fns->try_resume();
 	else
@@ -329,14 +272,16 @@ int toi_lowlevel_builtin(void)
 	}
 
 	restore_processor_state();
+	touch_softlockup_watchdog();
 
 	return error;
 }
-
 EXPORT_SYMBOL_GPL(toi_lowlevel_builtin);
 
-unsigned long toi_compress_bytes_in, toi_compress_bytes_out;
+unsigned long toi_compress_bytes_in;
 EXPORT_SYMBOL_GPL(toi_compress_bytes_in);
+
+unsigned long toi_compress_bytes_out;
 EXPORT_SYMBOL_GPL(toi_compress_bytes_out);
 
 unsigned long toi_state = ((1 << TOI_BOOT_TIME) |
@@ -351,25 +296,36 @@ EXPORT_SYMBOL_GPL(nr_hibernates);
 int toi_running;
 EXPORT_SYMBOL_GPL(toi_running);
 
-int toi_in_hibernate __nosavedata;
+__nosavedata int toi_in_hibernate;
 EXPORT_SYMBOL_GPL(toi_in_hibernate);
 
 __nosavedata struct pbe *restore_highmem_pblist;
-
-#ifdef CONFIG_TOI_CORE_EXPORTS
-#ifdef CONFIG_HIGHMEM
-EXPORT_SYMBOL_GPL(nr_free_highpages);
-EXPORT_SYMBOL_GPL(saveable_highmem_page);
 EXPORT_SYMBOL_GPL(restore_highmem_pblist);
-#endif
-#endif
 
-#if defined(CONFIG_TOI_CORE_EXPORTS) || defined(CONFIG_TOI_PAGEFLAGS_EXPORTS)
-EXPORT_SYMBOL_GPL(max_pfn);
-#endif
+int toi_trace_allocs;
+EXPORT_SYMBOL_GPL(toi_trace_allocs);
 
-#if defined(CONFIG_TOI_EXPORTS) || defined(CONFIG_TOI_CORE_EXPORTS)
-EXPORT_SYMBOL_GPL(snprintf_used);
+void toi_read_lock_tasklist(void)
+{
+	read_lock(&tasklist_lock);
+}
+EXPORT_SYMBOL_GPL(toi_read_lock_tasklist);
+
+void toi_read_unlock_tasklist(void)
+{
+	read_unlock(&tasklist_lock);
+}
+EXPORT_SYMBOL_GPL(toi_read_unlock_tasklist);
+
+#ifdef CONFIG_TOI_ZRAM_SUPPORT
+int (*toi_flag_zram_disks) (void);
+EXPORT_SYMBOL_GPL(toi_flag_zram_disks);
+
+int toi_do_flag_zram_disks(void)
+{
+	return toi_flag_zram_disks ? (*toi_flag_zram_disks)() : 0;
+}
+EXPORT_SYMBOL_GPL(toi_do_flag_zram_disks);
 #endif
 
 static int __init toi_wait_setup(char *str)
@@ -388,3 +344,101 @@ static int __init toi_wait_setup(char *str)
 }
 
 __setup("toi_wait", toi_wait_setup);
+
+static int __init toi_translate_retry_setup(char *str)
+{
+	toi_translate_err_default = 0;
+	return 1;
+}
+
+__setup("toi_translate_retry", toi_translate_retry_setup);
+
+static int __init toi_debug_setup(char *str)
+{
+	toi_bkd.toi_action |= (1 << TOI_LOGALL);
+	toi_bootflags_mask |= (1 << TOI_LOGALL);
+	toi_bkd.toi_debug_state = 255;
+	toi_bkd.toi_default_console_level = 7;
+	return 1;
+}
+
+__setup("toi_debug_setup", toi_debug_setup);
+
+static int __init toi_pause_setup(char *str)
+{
+	toi_bkd.toi_action |= (1 << TOI_PAUSE);
+	toi_bootflags_mask |= (1 << TOI_PAUSE);
+	return 1;
+}
+
+__setup("toi_pause", toi_pause_setup);
+
+#ifdef CONFIG_PM_DEBUG
+static int __init toi_trace_allocs_setup(char *str)
+{
+	int value;
+
+	if (sscanf(str, "=%d", &value))
+		toi_trace_allocs = value;
+
+	return 1;
+}
+__setup("toi_trace_allocs", toi_trace_allocs_setup);
+#endif
+
+static int __init toi_ignore_late_initcall_setup(char *str)
+{
+	int value;
+
+	if (sscanf(str, "=%d", &value))
+		ignore_late_initcall = value;
+
+	return 1;
+}
+
+__setup("toi_initramfs_resume_only", toi_ignore_late_initcall_setup);
+
+static int __init toi_force_no_multithreaded_setup(char *str)
+{
+	int value;
+
+	toi_bkd.toi_action &= ~(1 << TOI_NO_MULTITHREADED_IO);
+	toi_bootflags_mask |= (1 << TOI_NO_MULTITHREADED_IO);
+
+	if (sscanf(str, "=%d", &value) && value)
+		toi_bkd.toi_action |= (1 << TOI_NO_MULTITHREADED_IO);
+
+	return 1;
+}
+
+__setup("toi_no_multithreaded", toi_force_no_multithreaded_setup);
+
+#ifdef CONFIG_KGDB
+static int __init toi_post_resume_breakpoint_setup(char *str)
+{
+	int value;
+
+	toi_bkd.toi_action &= ~(1 << TOI_POST_RESUME_BREAKPOINT);
+	toi_bootflags_mask |= (1 << TOI_POST_RESUME_BREAKPOINT);
+	if (sscanf(str, "=%d", &value) && value)
+		toi_bkd.toi_action |= (1 << TOI_POST_RESUME_BREAKPOINT);
+
+	return 1;
+}
+
+__setup("toi_post_resume_break", toi_post_resume_breakpoint_setup);
+#endif
+
+static int __init toi_disable_readahead_setup(char *str)
+{
+	int value;
+
+	toi_bkd.toi_action &= ~(1 << TOI_NO_READAHEAD);
+	toi_bootflags_mask |= (1 << TOI_NO_READAHEAD);
+	if (sscanf(str, "=%d", &value) && value)
+		toi_bkd.toi_action |= (1 << TOI_NO_READAHEAD);
+
+	return 1;
+}
+
+__setup("toi_no_readahead", toi_disable_readahead_setup);
