@@ -138,6 +138,7 @@ static char *result_strings[] = {
 	"Failed to disable usermode helpers",
 	"Can't resume from alternate image",
 	"Header reservation too small",
+	"Device Power Management Preparation failed",
 };
 
 /**
@@ -829,6 +830,12 @@ static int do_prepare_image(void)
 
 	trap_non_toi_io = 1;
 
+	if (dpm_prepare(PMSG_FREEZE)) {
+		set_abort_result(TOI_DPM_PREPARE_FAILED);
+		dpm_complete(PMSG_FREEZE);
+		return 1;
+	}
+
 	pm_restrict_gfp_mask();
 
 	return 0;
