@@ -1,7 +1,7 @@
 /*
  * kernel/power/tuxonice_netlink.c
  *
- * Copyright (C) 2004-2007 Nigel Cunningham (nigel at tuxonice net)
+ * Copyright (C) 2004-2010 Nigel Cunningham (nigel at tuxonice net)
  *
  * This file is released under the GPLv2.
  *
@@ -10,12 +10,14 @@
 
 
 #include <linux/suspend.h>
+#include <linux/sched.h>
 #include "tuxonice_netlink.h"
 #include "tuxonice.h"
 #include "tuxonice_modules.h"
 #include "tuxonice_alloc.h"
+#include "tuxonice_builtin.h"
 
-struct user_helper_data *uhd_list;
+static struct user_helper_data *uhd_list;
 
 /*
  * Refill our pool of SKBs for use in emergencies (eg, when eating memory and
@@ -78,6 +80,10 @@ void toi_send_netlink_message(struct user_helper_data *uhd,
 	if (uhd->pid == -1)
 		return;
 
+	if (uhd->debug)
+		printk(KERN_ERR "toi_send_netlink_message: Send "
+				"message type %d.\n", type);
+
 	skb = toi_get_skb(uhd);
 	if (!skb) {
 		printk(KERN_INFO "toi_netlink: Can't allocate skb!\n");
@@ -94,17 +100,17 @@ void toi_send_netlink_message(struct user_helper_data *uhd,
 
 	netlink_unicast(uhd->nl, skb, uhd->pid, 0);
 
-	read_lock(&tasklist_lock);
-	t = find_task_by_pid(uhd->pid);
+	toi_read_lock_tasklist();
+	t = find_task_by_pid_ns(uhd->pid, &init_pid_ns);
 	if (!t) {
-		read_unlock(&tasklist_lock);
+		toi_read_unlock_tasklist();
 		if (uhd->pid > -1)
 			printk(KERN_INFO "Hmm. Can't find the userspace task"
 				" %d.\n", uhd->pid);
 		return;
 	}
 	wake_up_process(t);
-	read_unlock(&tasklist_lock);
+	toi_read_unlock_tasklist();
 
 	yield();
 
@@ -113,29 +119,36 @@ void toi_send_netlink_message(struct user_helper_data *uhd,
 nlmsg_failure:
 	if (skb)
 		put_skb(uhd, skb);
+
+	if (uhd->debug)
+		printk(KERN_ERR "toi_send_netlink_message: Failed to send "
+				"message type %d.\n", type);
 }
 EXPORT_SYMBOL_GPL(toi_send_netlink_message);
 
 static void send_whether_debugging(struct user_helper_data *uhd)
 {
-	static int is_debugging = 1;
+	static u8 is_debugging = 1;
 
 	toi_send_netlink_message(uhd, NETLINK_MSG_IS_DEBUGGING,
-			&is_debugging, sizeof(int));
+			&is_debugging, sizeof(u8));
 }
 
 /*
  * Set the PF_NOFREEZE flag on the given process to ensure it can run whilst we
  * are hibernating.
  */
-static int nl_set_nofreeze(struct user_helper_data *uhd, int pid)
+static int nl_set_nofreeze(struct user_helper_data *uhd, __u32 pid)
 {
 	struct task_struct *t;
 
-	read_lock(&tasklist_lock);
-	t = find_task_by_pid(pid);
+	if (uhd->debug)
+		printk(KERN_ERR "nl_set_nofreeze for pid %d.\n", pid);
+
+	toi_read_lock_tasklist();
+	t = find_task_by_pid_ns(pid, &init_pid_ns);
 	if (!t) {
-		read_unlock(&tasklist_lock);
+		toi_read_unlock_tasklist();
 		printk(KERN_INFO "Strange. Can't find the userspace task %d.\n",
 				pid);
 		return -EINVAL;
@@ -143,7 +156,7 @@ static int nl_set_nofreeze(struct user_helper_data *uhd, int pid)
 
 	t->flags |= PF_NOFREEZE;
 
-	read_unlock(&tasklist_lock);
+	toi_read_unlock_tasklist();
 	uhd->pid = pid;
 
 	toi_send_netlink_message(uhd, NETLINK_MSG_NOFREEZE_ACK, NULL, 0);
@@ -154,12 +167,13 @@ static int nl_set_nofreeze(struct user_helper_data *uhd, int pid)
 /*
  * Called when the userspace process has informed us that it's ready to roll.
  */
-static int nl_ready(struct user_helper_data *uhd, int version)
+static int nl_ready(struct user_helper_data *uhd, u32 version)
 {
 	if (version != uhd->interface_version) {
 		printk(KERN_INFO "%s userspace process using invalid interface"
-				" version. Trying to continue without it.\n",
-				uhd->name);
+				" version (%d - kernel wants %d). Trying to "
+				"continue without it.\n",
+				uhd->name, version, uhd->interface_version);
 		if (uhd->not_ready)
 			uhd->not_ready();
 		return -EINVAL;
@@ -173,7 +187,7 @@ static int nl_ready(struct user_helper_data *uhd, int version)
 void toi_netlink_close_complete(struct user_helper_data *uhd)
 {
 	if (uhd->nl) {
-		sock_release(uhd->nl->sk_socket);
+		netlink_kernel_release(uhd->nl);
 		uhd->nl = NULL;
 	}
 
@@ -185,29 +199,32 @@ void toi_netlink_close_complete(struct user_helper_data *uhd)
 
 	uhd->pid = -1;
 }
+EXPORT_SYMBOL_GPL(toi_netlink_close_complete);
 
 static int toi_nl_gen_rcv_msg(struct user_helper_data *uhd,
 		struct sk_buff *skb, struct nlmsghdr *nlh)
 {
-	int type;
+	int type = nlh->nlmsg_type;
 	int *data;
 	int err;
 
+	if (uhd->debug)
+		printk(KERN_ERR "toi_user_rcv_skb: Received message %d.\n",
+				type);
+
 	/* Let the more specific handler go first. It returns
 	 * 1 for valid messages that it doesn't know. */
 	err = uhd->rcv_msg(skb, nlh);
 	if (err != 1)
 		return err;
 
-	type = nlh->nlmsg_type;
-
 	/* Only allow one task to receive NOFREEZE privileges */
 	if (type == NETLINK_MSG_NOFREEZE_ME && uhd->pid != -1) {
-		printk("Received extra nofreeze me requests.\n");
+		printk(KERN_INFO "Received extra nofreeze me requests.\n");
 		return -EBUSY;
 	}
 
-	data = (int *)NLMSG_DATA(nlh);
+	data = NLMSG_DATA(nlh);
 
 	switch (type) {
 	case NETLINK_MSG_NOFREEZE_ME:
@@ -216,11 +233,13 @@ static int toi_nl_gen_rcv_msg(struct user_helper_data *uhd,
 		send_whether_debugging(uhd);
 		return 0;
 	case NETLINK_MSG_READY:
-		if (nlh->nlmsg_len < NLMSG_LENGTH(sizeof(int))) {
+		if (nlh->nlmsg_len != NLMSG_LENGTH(sizeof(u32))) {
 			printk(KERN_INFO "Invalid ready mesage.\n");
+			if (uhd->not_ready)
+				uhd->not_ready();
 			return -EINVAL;
 		}
-		return nl_ready(uhd, *data);
+		return nl_ready(uhd, (u32) *data);
 	case NETLINK_MSG_CLEANUP:
 		toi_netlink_close_complete(uhd);
 		return 0;
@@ -229,11 +248,17 @@ static int toi_nl_gen_rcv_msg(struct user_helper_data *uhd,
 	return -EINVAL;
 }
 
-static void toi_user_rcv_skb(struct user_helper_data *uhd,
-				  struct sk_buff *skb)
+static void toi_user_rcv_skb(struct sk_buff *skb)
 {
 	int err;
 	struct nlmsghdr *nlh;
+	struct user_helper_data *uhd = uhd_list;
+
+	while (uhd && uhd->netlink_id != skb->sk->sk_protocol)
+		uhd = uhd->next;
+
+	if (!uhd)
+		return;
 
 	while (skb->len >= NLMSG_SPACE(0)) {
 		u32 rlen;
@@ -255,30 +280,14 @@ static void toi_user_rcv_skb(struct user_helper_data *uhd,
 	}
 }
 
-static void toi_netlink_input(struct sock *sk, int len)
-{
-	struct user_helper_data *uhd = uhd_list;
-
-	while (uhd && uhd->netlink_id != sk->sk_protocol)
-		uhd = uhd->next;
-
-	do {
-		struct sk_buff *skb;
-		while ((skb = skb_dequeue(&sk->sk_receive_queue)) != NULL) {
-			toi_user_rcv_skb(uhd, skb);
-			put_skb(uhd, skb);
-		}
-	} while (uhd->nl && uhd->nl->sk_receive_queue.qlen);
-}
-
 static int netlink_prepare(struct user_helper_data *uhd)
 {
 	uhd->next = uhd_list;
 	uhd_list = uhd;
 
 	uhd->sock_seq = 0x42c0ffee;
-	uhd->nl = netlink_kernel_create(uhd->netlink_id, 0,
-			toi_netlink_input, NULL, THIS_MODULE);
+	uhd->nl = netlink_kernel_create(&init_net, uhd->netlink_id, 0,
+			toi_user_rcv_skb, NULL, THIS_MODULE);
 	if (!uhd->nl) {
 		printk(KERN_INFO "Failed to allocate netlink socket for %s.\n",
 				uhd->name);
@@ -294,11 +303,11 @@ void toi_netlink_close(struct user_helper_data *uhd)
 {
 	struct task_struct *t;
 
-	read_lock(&tasklist_lock);
-	t = find_task_by_pid(uhd->pid);
+	toi_read_lock_tasklist();
+	t = find_task_by_pid_ns(uhd->pid, &init_pid_ns);
 	if (t)
 		t->flags &= ~PF_NOFREEZE;
-	read_unlock(&tasklist_lock);
+	toi_read_unlock_tasklist();
 
 	toi_send_netlink_message(uhd, NETLINK_MSG_CLEANUP, NULL, 0);
 }
@@ -306,13 +315,16 @@ EXPORT_SYMBOL_GPL(toi_netlink_close);
 
 int toi_netlink_setup(struct user_helper_data *uhd)
 {
+	/* In case userui didn't cleanup properly on us */
+	toi_netlink_close_complete(uhd);
+
 	if (netlink_prepare(uhd) < 0) {
 		printk(KERN_INFO "Netlink prepare failed.\n");
 		return 1;
 	}
 
 	if (toi_launch_userspace_program(uhd->program, uhd->netlink_id,
-				UMH_NO_WAIT) < 0) {
+				UMH_WAIT_EXEC, uhd->debug) < 0) {
 		printk(KERN_INFO "Launch userspace program failed.\n");
 		toi_netlink_close_complete(uhd);
 		return 1;
